#include <windows.h>
#include "../../core/version.h"

//CREATEPROCESS_MANIFEST_RESOURCE_ID RT_MANIFEST "Application.manifest"

IDI_MYICON ICON   "punes.ico"

VS_VERSION_INFO VERSIONINFO
FILEVERSION     VER1_INT,VER2_INT
PRODUCTVERSION  VER1_INT,VER2_INT
FILEOS          VOS__WINDOWS32
FILETYPE        VFT_APP
BEGIN
	BLOCK "StringFileInfo"
	BEGIN
		BLOCK "040904E4"
		BEGIN
			VALUE "CompanyName",      AUTHOR
			VALUE "FileDescription",  "Nintendo Entertainment System Emulator"
			VALUE "FileVersion",      VER1 "." VER2
			VALUE "InternalName",     NAME
			VALUE "LegalCopyright",   COPYRANSI
			VALUE "LegalTrademarks",  AUTHOR
			VALUE "OriginalFilename", NAME ".exe"
			VALUE "ProductName",      NAME
			VALUE "ProductVersion",   VER1 "." VER2
		END
	END

	BLOCK "VarFileInfo"
	BEGIN
		VALUE "Translation", 0x0409, 1252
	END
END
