/*
 *  Copyright (C) 2010-2024 Fabio Cavallo (aka FHorse)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "wdgMenuBar.hpp"
#include "mainWindow.hpp"
#include "gui.h"
#include "conf.h"

wdgMenuBar::wdgMenuBar(QWidget *parent) : QMenuBar(parent) {}
wdgMenuBar::~wdgMenuBar() = default;

void wdgMenuBar::mouseDoubleClickEvent(QMouseEvent *event) {
	if (!event->spontaneous() || ((cfg->fullscreen == FULLSCR) && (gfx.type_of_fscreen_in_use != FULLSCR_IN_WINDOW))) {
		return;
	}

	if (event->button() == Qt::LeftButton) {
		mainwin->toggle_toolbars();
		cfg->toolbar.hidden = mainwin->toolbar->isHidden();
	} else if (event->button() == Qt::RightButton) {
		mainwin->action_Toggle_GUI_in_window->trigger();
	}
}
