"#define PI 3.14159265\n"
"\n"
"#if defined(TWO_PHASE)\n"
"#define CHROMA_MOD_FREQ (4.0 * PI / 15.0)\n"
"#elif defined(THREE_PHASE)\n"
"#define CHROMA_MOD_FREQ (PI / 3.0)\n"
"#endif\n"
"\n"
"#if defined(COMPOSITE)\n"
"#define SATURATION 1.0\n"
"#define BRIGHTNESS 1.0\n"
"#define ARTIFACTING 1.0\n"
"#define FRINGING 1.0\n"
"#elif defined(SVIDEO)\n"
"#define SATURATION 1.0\n"
"#define BRIGHTNESS 1.0\n"
"#define ARTIFACTING 0.0\n"
"#define FRINGING 0.0\n"
"#endif\n"
"\n"
"#if defined(COMPOSITE) || defined(SVIDEO)\n"
"const float3x3 mix_mat = float3x3(\n"
"      BRIGHTNESS, ARTIFACTING, ARTIFACTING,\n"
"      FRINGING, 2.0 * SATURATION, 0.0,\n"
"      FRINGING, 0.0, 2.0 * SATURATION \n"
");\n"
"#endif\n"
"\n"

