{
"#if defined(VERTEX)\n"
"attribute vec2 VertexCoord;\n"
"attribute vec4 Color;\n"
"attribute vec2 TexCoord;\n"
"uniform mat4 MVPMatrix;\n"
"varying vec2 tex_coord;\n"
"varying vec4 color;\n"
"void main() {\n"
"   gl_Position = MVPMatrix * vec4(VertexCoord, 0.0, 1.0);\n"
"   tex_coord = TexCoord;\n"
"   color = Color;\n"
"}\n"
"#elif defined(FRAGMENT)\n"
"uniform sampler2D Texture;\n"
"varying vec2 tex_coord;\n"
"varying vec4 color;\n"
"void main() {\n"
"   gl_FragColor = color * texture2D(Texture, tex_coord);\n"
"}\n"
"#endif\n"
},

