{
#include "ntsc-pass2-vertex.inc"
#include "ntsc-decode-filter-2phase.inc"
#include "ntsc-rgbyuv.inc"
"\n"
"#define fetch_offset(offset, one_x) \\\n"
"   tex2D(s0, vertex.tex + float2((offset) * (one_x), 0.0)).xyz\n"
"\n"
"float4 main_fragment (uniform input IN, in data vertex, uniform sampler2D s0 : TEXUNIT0) : COLOR\n"
"{\n"
#include "ntsc-pass2-decode.inc"
"   float3 rgb = yiq2rgb(signal);\n"
"   return float4(rgb, 1.0);\n"
"}\n"
"\n"
},
