"struct data\n"
"{\n"
"   float2 tex;\n"
"   float2 pix_no;\n"
"};\n"
"\n"
"struct input\n"
"{\n"
"   float2 video_size;\n"
"   float2 texture_size;\n"
"   float2 output_size;\n"
"   float frame_count;\n"
"};\n"
"\n"
"void main_vertex\n"
"(\n"
"   float4 position : POSITION,\n"
"   out float4 oPosition : POSITION,\n"
"   uniform float4x4 modelViewProj,\n"
"   float2 tex : TEXCOORD,\n"
"   uniform input IN,\n"
"   out data oData\n"
")\n"
"{\n"
"   oPosition = mul(modelViewProj, position);\n"
"   oData.tex = tex;\n"
"   oData.pix_no = tex * IN.texture_size * (IN.output_size / IN.video_size);\n"
"}\n"
"\n"
"\n"
